!+

! Module: omfastflat_mod
!
! Authors:  Jon Rainger + Tim Harries - MSSL
!          
! Description: Flatfields fast mode images
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
! History:
!          1999 Jan: updated psw - spr 553
!          2001 May 08 (V.Y.): modificated slewFlat array correction for the flat field 
!                              (eliminated windowX0P(0) and windowY0P(0) from the indexes
!                               because rawXP and rawYP alreay containes the origin of the window)
!          2001 May 08 (V.Y.)  filling in the fastImgP array in the correct way 
!          2001 October 17 (VNY) - data set in the /test directory has been updated
!          2001 October 19 (VNY) - more robust call to CheckArrayBounds subroutine
!          2001 October 22 (VNY) - a claryfying message has been added to CheckArrayBounds subroutine
!          2001 October 23 (VNY) - additional checks and integer conversions to CheckArrayBounds
!                               to prevent a falure at Cosine3 machine
!          2001 October 24 (VNY) - pointers to window parameters (windowX0P, windowY0P, windowDXP, windowDYP)
!                                replaced with int32 variables in the cases when these parameters are 
!                                passed to inner subroutines (in order to avoid possible subscripts)
!          2001 October 25 (VNY) a bug with the use of factorP pointer has been fixed to rectify 
!                                possible test-harness in omdetect task 
!          2002 January 08 (VNY) DEPEND file has been updated to use higher version of ssclib (2.0)

module omfastflat_mod

  use dal
  use cal
  use param
  use errorHandling
  use copyTable_mod
  use sasdal

  implicit none

  private

  public omfastflat

contains

  !---------------------------------------------------------------------
  subroutine omfastflat
    !---------------------------------------------------------------------

    character(len=255) :: set                                         ! OM Fast Mode Event List filename (output of OMFASTSHIFT).
    character(len=255) :: slewFlatSet                                 ! OM Inorbit Flatfield filename (output of OMFLATGEN).
    character(len=255) :: fastImgSet                                  ! Flatfielded OM Fast Mode Image filename.
    character(len=255) :: OSWFlatSet                                  ! Data product file: OM OSW flat field filename. 

    type(DataSetT) :: inDataSetH                                      ! Input dataset handle.
    type(DataSetT) :: slewFlatDataSetH                                ! OM Inorbit Flatfield dataset handle.
    type(DataSetT) :: fastImgDataSetH                                 ! OM Fast Mode Image dataset handle.
    type(DataSetT) :: OSWFlatDataSetH                                 ! OM OSW flat field dataset handle.

    type(ArrayT) :: slewFlatArrayH                                    ! Input array handle.
    type(ArrayT) :: fastImgArrayH                                     ! Output array handle.
    type(ArrayT) :: qualityArrayH                                     ! Output array handle.
    type(ArrayT) :: OSWFlatArrayH                                     ! Output array handle.
    type(TableT) :: inTableH                                          ! Input table handle.
    type(TableT) :: modesTableH                                       ! Modes table dataset handle.

    type(ColumnT), dimension(5) :: inColumnH                          ! Input column handle.
    type(ColumnT), dimension(4) :: modesColumnH                       ! Modes column handle.
    type(ColumnT), dimension(2) :: outColumnH                         ! Output column handle.

    integer(kind=int16), dimension(:), pointer :: rawXP               ! Raw x column pointer.
    integer(kind=int16), dimension(:), pointer :: rawYP               ! Raw y column pointer.
    integer(kind=int16), dimension(:), pointer :: corrXP              ! Raw x column pointer.
    integer(kind=int16), dimension(:), pointer :: corrYP              ! Raw y column pointer.
    integer(kind=int8), dimension(:), pointer :: winFlagP             ! WinFlag pointer
    integer(kind=int32), dimension(:), pointer :: windowX0P           ! WindowX0 pointer.
    integer(kind=int32), dimension(:), pointer :: windowY0P           ! WindowY0 pointer.
    integer(kind=int32), dimension(:), pointer :: windowDXP           ! WindowDX pointer.
    integer(kind=int32), dimension(:), pointer :: windowDYP           ! WindowDY pointer.
!---VNY 24.10.2001
    integer(kind=int32) :: windowX0, windowY0, windowDX, windowDY

    real(kind=single), dimension(:, :), allocatable :: slewFlat       ! Real array pointer.
    real(kind=single), dimension(:, :), pointer :: slewFlatP          ! Real array pointer.

    integer(kind=int16), dimension(:, :), pointer :: badPixMapP       ! Integer array pointer to bad pixel map.   

    integer(kind=int8), dimension(:), pointer :: badPixFlagP          ! Bad pixel flag column pointer.
    real(kind=single), dimension(:), pointer :: factorP               ! QE Factor column pointer.

    integer(kind=int8), dimension(:, :), pointer :: qualityP          ! Quality array pointer.
    real(kind=single), dimension(:, :), pointer :: fastImgP           ! Real array pointer.

    real(kind=single), dimension(:, :), pointer :: OSWFlatImgP        ! OSW flat field array pointer.

    integer(kind=int32), dimension(2) :: imageDim                     ! Image dimension.
    integer(kind=int32) :: nEvents                                    ! Number of events.
    integer(kind=int32) :: i, i1                                          ! Counter.
    integer(kind=int32) :: j, j1                                          ! Counter.
!----------------Added by VY 8.05.2001
    integer(kind=int32) :: indexX                                     ! for calculation of the output image array index  
    integer(kind=int32) :: indexY
    integer :: intRawX, intRawY 
!------------------
    character(len=255) :: dataMode                                    ! Input file data mode.
    character(len=255) :: binBPE                                      ! Input file BPE binning keyword.
    character(len=255) :: filterMode                                  ! Input file filter mode

    ! get task parameters.

    set = stringParameter("set")
    slewFlatSet = stringParameter("slewflatset")
    fastImgSet = stringParameter("fastimgset")
    OSWFlatSet = stringParameter("oswflatset")


    ! Check data types of ODF input files

    ! **** JFR **** need to check if input is an Inorbit flatfield.

    ! Get a hold on the Inorbit flat field File.


    slewFlatDataSetH = dataSet(slewFlatSet, read)

    slewFlatArrayH = array(slewFlatDataSetH, "PRIMARY", Read)
    slewFlatP => real32Array2Data(slewFlatArrayH)


    ! Get a hold on the ODF OM Fast Mode Event List File.

    inDataSetH = dataSet(set, modify)

    ! Determine file type from file name in primary header.


    dataMode = trim(stringAttribute(inDataSetH, "DATAMODE")) 

    ! Determine filter type from the input data
    
    filterMode = trim(stringAttribute(inDataSetH, "FILTER"))


    ! Determine if BPE binning is on/off

    binBPE = trim(stringAttribute(inDataSetH, "BINBPE")) 

    if (dataMode == "FAST") then

       ! Get a handle on the tracking history data binary table extension.

       inTableH = table(inDataSetH, "OMFAE1")

       ! Get handles and pointers on the event list binary table extension columns.

       inColumnH(1) = column(inTableH, "RAWX", read)
       rawXP => int16Data(inColumnH(1))

       inColumnH(2) = column(inTableH, "RAWY", read)
       rawYP => int16Data(inColumnH(2))

       inColumnH(3) = column(inTableH, "CORR_X", read)
       corrXP => int16Data(inColumnH(3))

       inColumnH(4) = column(inTableH, "CORR_Y", read)
       corrYP => int16Data(inColumnH(4))

       inColumnH(5) = column(inTableH, "WIN_FLAG", Read)
       winFlagP => int8data(inColumnH(5))


       ! Get the number of rows in the event list binary table extension.

       nEvents = numberOfRows(inTableH)

       ! Get window information from modes extension.

       modesTableH = table(inDataSetH, "MODES")

       modesColumnH(1) = column(modesTableH, "WINDOWX0", read)
       windowX0P => int32data(modesColumnH(1))

       modesColumnH(2) = column(modesTableH, "WINDOWY0", read)
       windowY0P => int32data(modesColumnH(2))

       modesColumnH(3) = column(modesTableH, "WINDOWDX", read)
       windowDXP => int32data(modesColumnH(3))

       modesColumnH(4) = column(modesTableH, "WINDOWDY", read)
       windowDYP => int32data(modesColumnH(4))

!---VNY 24.10.2001
       windowX0 = windowX0P(0)
       windowY0 = windowY0P(0)
       windowDX = windowDXP(0)
       windowDY = windowDYP(0)
       
    else

       ! Incompatible file.

       call fatal("badFileMode","Value for keyword DATAMODE is incompatible")

    end if


    ! Set the state of the cal.

    call CAL_setState(OM, 1, 0)

    ! Get the bad pixel map.

    call CAL_getBadPixelMap(badPixMapP)

    ! Add factor and bad pixel columns to event list file.

    outColumnH(1) = addColumn(inTableH, "FACTOR", real32)
    factorP => real32data(outColumnH(1))

    outColumnH(2) = addColumn(inTableH, "BADPIX", integer8)
    badPixFlagP => int8data(outColumnH(2))

    ! Initialise pointer arrays to zero.

    factorP = 0.0
    badPixFlagP = 0

    ! Calculate factor and bad pixel flag.

    ! If BPE binning off then need to rebin flatfield.


    if (binBPE == "T") then
       allocate(slewFlat(0:1023,0:1023))
       slewFlat = slewFlatP
    else 

       allocate(slewFlat(0:2047,0:2047))       
       do i= 1, 1024
          do j = 1, 1024
!--- VNY 23.10.2001 Additional conversions for Cosine3 machine
             intRawX=2*i-2
             intRawY=2*j-2
             call CheckArrayBounds(slewFlat, intRawX,intRawY )
             intRawX=2*i-1
             intRawY=2*j-1
             call CheckArrayBounds(slewFlat, intRawX,intRawY)
             intRawX=2*i-2
             intRawY=2*j-1
             call CheckArrayBounds(slewFlat, intRawX, intRawY)
             intRawX=2*i-1
             intRawY=2*j-2
             call CheckArrayBounds(slewFlat, intRawX, intRawY)

             slewFlat(2*i-2,2*j-2) = slewflatP(i-1,j-1)
             slewFlat(2*i-1,2*j-1) = slewflatP(i-1,j-1)
             slewFlat(2*i-2,2*j-1) = slewflatP(i-1,j-1)
             slewFlat(2*i-1,2*j-2) = slewflatP(i-1,j-1)
          end do
       end do
    end if


    do i = 0, nEvents - 1
       intRawX=rawXP(i)
       intRawY=rawYP(i)
!---VNY 23.10.2001 Additional check for Cosine3 machine
       if (intRawX .lt. 2048 .and. intRawY .lt. 2048) then
          call CheckArrayBounds(slewFlat, intRawX, intRawY)
       else
          if (intRawX .ge. 2048) then
             call warning("X exceeds 2047") 
             intRawX=2047
           end if
          if (intRawY .ge. 2048) then
             call warning("Y exceeds 2047") 
             intRawY=2047
           end if
       end if

       if(slewFlat(rawXP(i), rawYP(i)) /= 0) then
          factorP(i) = (slewFlat(rawXP(i), rawYP(i)))**(-1.0)
       else
          factorP(i) = 0
       end if

       intRawX=rawXP(i)
       intRawY=rawYP(i)
       call CheckInt16ArrayBounds(badPixMapP, intRawX, intRawY)

       if (badPixMapP(rawXP(i), rawYP(i)) /= 0) then
          badPixFlagP(i) = 1
       end if
    end do


    ! Create flatfielded fast mode image file.

    fastImgDataSetH = dataSet(fastImgSet, Create)

    ! Copy attributes.

    call copyAllAttributes(attributable(inDataSetH), attributable(fastImgDataSetH))

    !   call setAttribute(outDataSetH, "CONTENT", "OM OSW IMAGE", "Content of file")
    !   call setAttribute(OutDataSetH, "HDUCLASS", "OGIP", "Format conforms to OGIP/GSFC conventions")
    !   call setAttribute(OutDataSetH, "HDUCLAS1", "IMAGE", "File contains an image")
    !   call setAttribute(OutDataSetH, "HDUCLAS2", "TOTAL", "Total counts")

    imageDim(1) = windowDX
    imageDim(2) = windowDY

    ! Get image handle and pointer.

    fastImgArrayH = addArray(fastImgDataSetH, "PRIMARY", real32, ImageDim)
    fastImgP => real32array2data(fastImgArrayH)

    ! Get quality array handle and pointer.

    qualityArrayH = addArray(fastImgDataSetH, "QUALITY", integer8, ImageDim)
    qualityP => int8array2data(qualityArrayH)

    ! Build image and quality array.

    fastImgP = 0.0

!-------------------------Corrections by V.Y. 8.05.2001 (filling in the fastImgP arrayin a correct way)
    qualityP = 0
    do i = 0 , nEvents - 1


       if (winFlagP(i) /= 1) then          
          indexX=corrXP(i)-windowX0
          if (indexX<0) then
             indexX=0
          end if

          indexY=corrYP(i)-windowY0
          if (indexY<0) then
             indexY=0
          end if
          if (indexX .ge. 0 .and. indexX .lt. imageDim(1) .and. indexY .ge. 0 .and. indexY .lt. imageDim(2)) then
             fastImgP(indexX, indexY) = fastImgP(indexX, indexY) + factorP(i) 
             if (badPixFlagP(i) == 1) then
                qualityP(indexX, indexY) = 1
             end if
          end if
       end if
    end do

!---VNY 25.10.2001
    ! Release the cal memory.
    call cal_releaseMemory(badPixMapP)
    ! Release handles on data.
    call release(outColumnH(1))
    call release(outColumnH(2))


    call copyTable(inDataSetH, fastImgDataSetH, "MODES")

    ! Create OM OSW flat field data product.

    OSWFlatDataSetH = dataSet(OSWFlatSet, create)

    call copyAllAttributes(attributable(inDataSetH), attributable(OSWFlatDataSetH))

    call setAttribute(OSWFlatDataSetH, "CONTENT", "OM OSW FLAT FIELD", "Content of file")
    call setAttribute(OSWFlatDataSetH, "HDUCLASS", "OGIP", "Format conforms to OGIP/GSFC conventions")
    call setAttribute(OSWFlatDataSetH, "HDUCLAS1", "IMAGE", "File contains an image")
    call setAttribute(OSWFlatDataSetH, "HDUCLAS2", "DETMAP", "Detector map")

    OSWFlatArrayH = addArray(OSWFlatDataSetH, "PRIMARY", real32, imageDim)
    OSWFlatImgP => real32array2data(OSWFlatArrayH)

    do i=0,windowDX - 1
       do j=0,windowDY - 1
          intRawX=windowX0+i
          intRawY=windowY0+j

!---VNY 23.10.2001 Additional check for Cosine3 machine
          if (intRawX .lt. 2048 .and. intRawY .lt.2048) then
             call CheckArrayBounds(slewFlat, intRawX, intRawY)
          else
             if (intRawX .ge. 2048) then
                call warning("X exceeds 2047")
                intRawX=2047
             end if
             if (intRawY .ge.2048) then
                call warning("Y exceeds 2047")
                intRawY=2047
             end if
          end if
!---VNY 23.10.2001 Additional check for Cosine3 machine
          i1=i
          j1=j
          if (i1 .lt. 2048 .and. j1 .lt. 2048) then 
             call CheckArrayBounds(OSWFlatImgP, i1,j1)
          else
             if (i1 .ge. 2048) then
                call warning("I exceeds 2047")
                i1=2047
             end if
             if (j1 .ge. 2048) then
                call warning("J exceeds 2047")
                j1=2047
             end if
          end if
          OSWFlatImgP(i1,j1) = slewFlat(intRawX, intRawY)
       end do
    end do

    call copyTable(inDataSetH, OSWFlatDataSetH, "MODES")


    ! Free up memory and pointers.

    call release(modesColumnH(1))
    call release(modesColumnH(2))
    call release(modesColumnH(3))
    call release(modesColumnH(4))
    call release(modesTableH)

    call release(OSWFlatArrayH)
    call release(OSWFlatDataSetH)

    call release(qualityArrayH)
    call release(fastImgArrayH)
    call release(fastImgDataSetH)

    call release(inColumnH(1))
    call release(inColumnH(2))
    call release(inColumnH(3))
    call release(inColumnH(4))
    call release(inColumnH(5))
    call release(inTableH)

    call release(inDataSetH)
    call release(slewFlatArrayH)
    call release(slewFlatDataSetH)


  end subroutine omfastflat

subroutine CheckArrayBounds(array, i, j)
    real(kind=single), intent(in), dimension(: , :) :: array
    integer, intent(in) :: i, j
    integer :: iMax, jMax

    iMax = size(array, 1) - 1
    jMax = size(array, 2) - 1

    if (i .lt. 0) call fatal("RealArray bounds error- i < 0")
    if (j .lt. 0) call fatal("RealArray bounds error- j < 0")

    if (i .gt. iMax) then
       call fatal("RealArray bounds error- i > imax")
    end if

    if (j .gt. jMax) then
       call fatal("RealArray bounds error- j > jmax")
    end if

end subroutine CheckArrayBounds
subroutine CheckInt16ArrayBounds(array, i, j)
    integer(kind=int16), intent(in), dimension(: , :) :: array
    integer, intent(in) :: i, j
    integer :: iMax, jMax

    iMax = size(array, 1) - 1
    jMax = size(array, 2) - 1

    if (i .lt. 0) call fatal("Int16Array bounds error- i < 0")
    if (j .lt. 0) call fatal("Int16Array bounds error- j < 0")
    if (i .gt. iMax) call fatal("Int16Array bounds error- i > imax")
    if (j .gt. jMax) call fatal("Int16Array bounds error- j > jmax")

end subroutine CheckInt16ArrayBounds


end module omfastflat_mod






